#include "GesamtNetz.h"
//#include "String.h"

#define SQR(x)   ((x)*(x))


// Fuer allgemeine Ausgabe und Eingabe
// Format ist jetzt: [magicNumber]\n
//                   [Expertennetz Aouei]
//                   [Expertennetz Ouvoid]
//                   [Expertennetz Eivoid]
//                   [myStat]
ostream& operator<< (ostream &ostr, const GesamtNetz &source){
  if(((GesamtNetz &)source).getGueltigkeit()){
    ostr << source.magicNumber << endl;
    
    ostr << *(source.Aouei);
    ostr << *(source.Ouvoid);
    ostr << *(source.Eivoid);
    
    ostr << source.myStat;

    ostr << flush;
  }

  return ostr;
}



istream& operator>> (istream &istr, GesamtNetz &dest){
  dest.gueltigkeit = 0;

  istr >> dest.magicNumber[0];
  istr >> dest.magicNumber[1];
  istr >> dest.magicNumber[2];
  istr >> dest.magicNumber[3];
  dest.magicNumber[4] = 0;   // Laengebegrenzung

  istr >> *(dest.Aouei);
  istr >> *(dest.Ouvoid);
  istr >> *(dest.Eivoid);

  istr >> dest.myStat;

  // Wenn es bis hier gelungen ist, dann muss es gueltig sein!
  // (Das GesamtNetz wenigstens, aber nicht unbedingt die Expertennetze)
  dest.gueltigkeit = 1;

  return istr;
} 



GesamtNetz::GesamtNetz(fstream &fp):testFehler(0.0),trainingsFehler(0.0){
  Aouei = new Netz;
  Ouvoid = new Netz;
  Eivoid = new Netz;
  gueltigkeit = 0;

  strcpy(magicNumber,GESAMTNETZNOTMAGICNUMBER);

  fp >> *this;
}



GesamtNetz::GesamtNetz(int schichtenAnzahl1, int *tf1, int *neuronenAnzahl1,
		       int schichtenAnzahl2, int *tf2, int *neuronenAnzahl2,
		       int schichtenAnzahl3, int *tf3, int *neuronenAnzahl3,
		       int batchAnzahl, int trainingsfehlerAnzahl, 
		       int testfehlerAnzahl, double epsilon):testFehler(0.0),
							     trainingsFehler(0.0){
  Aouei = new Netz(schichtenAnzahl1, tf1, neuronenAnzahl1, batchAnzahl, trainingsfehlerAnzahl, testfehlerAnzahl, epsilon);
  Ouvoid = new Netz(schichtenAnzahl2, tf2, neuronenAnzahl2, batchAnzahl, trainingsfehlerAnzahl, testfehlerAnzahl, epsilon);
  Eivoid = new Netz(schichtenAnzahl3, tf3, neuronenAnzahl3, batchAnzahl, trainingsfehlerAnzahl, testfehlerAnzahl, epsilon);

  strcpy(magicNumber,GESAMTNETZMAGICNUMBER);

  gueltigkeit = 1;
}


// Leerer Konstruktor, damit man ein Gesamtnetz machen kann und spaeter
// durch >> einstellen.
GesamtNetz::GesamtNetz(void):testFehler(0.0),trainingsFehler(0.0){
  Aouei = new Netz;
  Ouvoid = new Netz;
  Eivoid = new Netz;
  gueltigkeit = 0;

  strcpy(magicNumber,GESAMTNETZNOTMAGICNUMBER);
}


bool GesamtNetz::speichern(fstream &fp){
  if (getGueltigkeit()) {
    fp << *this;
    return 1;
  } else
    return 0;
}


// der Benutzer muss selber FFTFenster delete[]n!
// Ich darf hier die Ergebnisse von Netz::trainiere() NICHT delete[]n
Vokal &GesamtNetz::trainiere(double *FFTFenster, char echterVokal){
  // AoueiSoll usw. entsprechen dem echterVokal:
  setSoll(echterVokal);

  // Propagation durchfuehren:
  AoueiOut = Aouei->trainiere(FFTFenster, AoueiSoll);
  OuvoidOut = Ouvoid->trainiere(FFTFenster, OuvoidSoll);
  EivoidOut = Eivoid->trainiere(FFTFenster, EivoidSoll);
  
  // Damit die Ausgaben wie Wahrscheinlichkeiten aussehen (Summe ist 1.0)
  // Und sie werden zu vok.Aouei[], vok.Ouvoid[] und vok.Eivoid[] kopiert.
  vomNetzenPseudonormieren();

  // Mit den Sollvektoren multiplizieren
  setSoll('a');
  vok.a = vokalWert();
  setSoll('e');
  vok.e = vokalWert();
  setSoll('i');
  vok.i = vokalWert();
  setSoll('o');
  vok.o = vokalWert();
  setSoll('u');
  vok.u = vokalWert();

  // Damit die Ausgabe wie Wahrscheinlichkeitswerte aussieht.
  outputPseudonormieren();

  // char vok.gewaehlt auswaehlen und vok.sicherheit zuordnen
  charWaehlen();

  // Statistik berechnen
  trainingsFehler += gesamtfehlerBerechnen(echterVokal);

  // Statistik updaten, falls es Zeit ist.
  bool refresh = Aouei->getTrainingsPlotRefresh();
  if (refresh) {
    // gesamtfehlerBerechnen macht: (1/2) Summe ((target-actual)^2)
    // gemittelt ueber traingsfehlerAnzahl
    myStat.addTraining(trainingsFehler/(double)Aouei->getTrainingsfehlerAnzahl());
    //cout << "GN trainingfehler: " << trainingsFehler/(double)Aouei->getTrainingsfehlerAnzahl() << endl;
    trainingsFehler = 0.0;
  }
    
  // Damit der Benutzer weiss, ob die Plots noch neu
  // gezeichnet werden sollen.
  vok.neuesFehlerplot = refresh;
  // es koennte eigentlich auch sein:
  // vok.neuesFehlerplot = (Aouei->getFehlerPlotRefresh() || 
  //                        Aouei->getTestPlotRefresh());
  vok.neueGewichtenwerte = (Aouei->getGewichtenPlotRefresh());
  //cout << "here!\n" << flush;

  return vok;
}


Vokal &GesamtNetz::test(double *FFTFenster, char echterVokal){
  // AoueiSoll usw. entsprechen dem echterVokal:
  setSoll(echterVokal);

  // Propagation durchfuehren:
  AoueiOut = Aouei->test(FFTFenster, AoueiSoll);
  OuvoidOut = Ouvoid->test(FFTFenster, OuvoidSoll);
  EivoidOut = Eivoid->test(FFTFenster, EivoidSoll);
  
  // Damit die Ausgaben wie Wahrscheinlichkeiten aussehen (Summe ist 1.0)
  // Und sie werden zu vok.Aouei[], vok.Ouvoid[] und vok.Eivoid[] kopiert.
  vomNetzenPseudonormieren();

  // Pseudowahrscheinlichkeiten fuer die fuenf Vokalen berechnen
  setSoll('a');
  vok.a = vokalWert();
  setSoll('e');
  vok.e = vokalWert();
  setSoll('i');
  vok.i = vokalWert();
  setSoll('o');
  vok.o = vokalWert();
  setSoll('u');
  vok.u = vokalWert();

  // Damit die Ausgabe wie Wahrscheinlichkeitswerte aussieht.
  outputPseudonormieren();

  // char vok.gewaehlt auswaehlen und vok.sicherheit zuordnen
  charWaehlen();

  // Statistik updaten, falls es Zeit ist
  testFehler += gesamtfehlerBerechnen(echterVokal);

  bool refresh = Aouei->getTestPlotRefresh();
  if (refresh) {
    // gesamtfehlerBerechnen macht: (1/2) * Summe ((target-actual)^2)
    // gemittelt ueber testfehlerAnzahl
    myStat.addTest(testFehler/(double)Aouei->getTestfehlerAnzahl());
    testFehler = 0.0;
  }
    
  // Damit der Benutzer weiss, ob die Plots noch neu
  // gezeichnet werden sollen.
  vok.neuesFehlerplot = refresh;
  // es koennte eigentlich auch sein:
  // vok.neuesFehlerplot = (Aouei->getFehlerPlotRefresh() || 
  //                        Aouei->getTestPlotRefresh());
  // Soll eigentlich immer 0 sein, aber um sicher zu sein:
  vok.neueGewichtenwerte = (Aouei->getGewichtenPlotRefresh());

  return vok;
}



Vokal &GesamtNetz::propagate(double *FFTFenster){
  // Propagation durchfuehren:
  AoueiOut = Aouei->propagate(FFTFenster);
  OuvoidOut = Ouvoid->propagate(FFTFenster);
  EivoidOut = Eivoid->propagate(FFTFenster);
  
  // Damit die Ausgaben wie Wahrscheinlichkeiten aussehen (Summe ist 1.0)
  // Und sie werden zu vok.Aouei[], vok.Ouvoid[] und vok.Eivoid[] kopiert.
  vomNetzenPseudonormieren();

  // Pseudowahrscheinlichkeiten fuer die fuenf Vokalen berechnen
  setSoll('a');
  vok.a = vokalWert();
  setSoll('e');
  vok.e = vokalWert();
  setSoll('i');
  vok.i = vokalWert();
  setSoll('o');
  vok.o = vokalWert();
  setSoll('u');
  vok.u = vokalWert();

  // Damit die Ausgabe wie normierte Wahrscheinlichkeitswerte aussieht.
  outputPseudonormieren();

  // char vok.gewaehlt auswaehlen und vok.sicherheit zuordnen
  charWaehlen();

  // Nichts haette geupdatet werden sollen, aber falls etwas doch
  // passiert ist:
  vok.neuesFehlerplot = (Aouei->getTrainingsPlotRefresh() || 
			 Aouei->getTestPlotRefresh());
  vok.neueGewichtenwerte = (Aouei->getGewichtenPlotRefresh());

  return vok;
}

void GesamtNetz::setSoll(char echterVokal){
  for(int i=0;i<3;i++){
    AoueiSoll[i] = 0.0;
    OuvoidSoll[i] = 0.0;
    EivoidSoll[i] = 0.0;
  }

  if (echterVokal=='a')
    AoueiSoll[0] = OuvoidSoll[2] = EivoidSoll[2] = 1.0;

  if (echterVokal=='e')
    AoueiSoll[2] = OuvoidSoll[2] = EivoidSoll[0] = 1.0;
  
  if (echterVokal=='i')
    AoueiSoll[2] = OuvoidSoll[2] = EivoidSoll[1] = 1.0;

  if (echterVokal=='o')
    AoueiSoll[1] = OuvoidSoll[0] = EivoidSoll[2] = 1.0;

  if (echterVokal=='u')
    AoueiSoll[1] = OuvoidSoll[1] = EivoidSoll[2] = 1.0;
}


// Eine kurze Hilfefunktion:
// (Falls die Transferfunktion der letzten Schicht Tanh(x)
// und deswegen auch negative Werte liefern koennte.
inline void nichtNeg(double *in){
  *in = fabs(*in);
  //if(*in < 0.0)
  //  *in = 0.0;
}

void GesamtNetz::vomNetzenPseudonormieren(void){
  double summe1=0.0, summe2=0.0, summe3=0.0;
  int i;

  for(i=0;i<3;i++){
    // Frag Oliver, ob sie wirklich zu Null gesetzt werden sollen.
    // Pseudowahrscheinlichkeiten duerften natuerlich nicht negativ sein.
    nichtNeg(&AoueiOut[i]);
    nichtNeg(&OuvoidOut[i]);
    nichtNeg(&EivoidOut[i]);

    summe1 += AoueiOut[i];
    summe2 += OuvoidOut[i];
    summe3 += EivoidOut[i];
  }
  for(i=0;i<3;i++){
    AoueiOut[i] /= summe1;
    OuvoidOut[i] /= summe2;
    EivoidOut[i] /= summe3;

    // Zur Ausgabevariable vok kopieren
    vok.Aouei[i] = AoueiOut[i];
    vok.Ouvoid[i] = OuvoidOut[i];
    vok.Eivoid[i] = EivoidOut[i];
  }
}

double GesamtNetz::vokalWert(void){
  double produkt=1.0;

  for(int i=0;i<3;i++){
    if (AoueiSoll[i]==1.0)
      produkt *= AoueiOut[i];
    if (OuvoidSoll[i]==1.0)
      produkt *= OuvoidOut[i];
    if (EivoidSoll[i]==1.0)
      produkt *= EivoidOut[i];
  }

  return produkt;
}

void GesamtNetz::outputPseudonormieren(void){
  // Frag Oliver, wie frueher
  nichtNeg(&vok.a);
  nichtNeg(&vok.e);
  nichtNeg(&vok.i);
  nichtNeg(&vok.o);
  nichtNeg(&vok.u);

  double summe = vok.a + vok.e + vok.i + vok.o + vok.u;

  vok.a /= summe;
  vok.e /= summe;
  vok.i /= summe;
  vok.o /= summe;
  vok.u /= summe;
}

void GesamtNetz::charWaehlen(void){
  vok.sicherheit = vok.a;
  vok.gewaehlt = 'a';
  if(vok.e > vok.sicherheit) {
    vok.sicherheit = vok.e;
    vok.gewaehlt = 'e';
  }
  if(vok.i > vok.sicherheit) {
    vok.sicherheit = vok.i;
    vok.gewaehlt = 'i';
  }
  if(vok.o > vok.sicherheit) {
    vok.sicherheit = vok.o;
    vok.gewaehlt = 'o';
  }
  if(vok.u > vok.sicherheit) {
    vok.sicherheit = vok.u;
    vok.gewaehlt = 'u';
  }
}

double GesamtNetz::gesamtfehlerBerechnen(char echterVokal){
  double tmp;
  switch (echterVokal) {
  case 'a':
    tmp = SQR(1.0-vok.a);
    tmp += SQR(vok.e) + SQR(vok.i) + SQR(vok.o) + SQR(vok.u);
    break;
  case 'e':
    tmp = SQR(1.0-vok.e);
    tmp += SQR(vok.a) + SQR(vok.i) + SQR(vok.o) + SQR(vok.u);
    break;
  case 'i':
    tmp = SQR(1.0-vok.i);
    tmp += SQR(vok.a) + SQR(vok.e) + SQR(vok.o) + SQR(vok.u);
    break;
  case 'o':
    tmp = SQR(1.0-vok.o);
    tmp += SQR(vok.a) + SQR(vok.e) + SQR(vok.i) + SQR(vok.u);
    break;
  case 'u':
    tmp = SQR(1.0-vok.u);
    tmp += SQR(vok.a) + SQR(vok.e) + SQR(vok.i) + SQR(vok.o);
    break;
  default:
    // dieser Fall darf eigentlich nicht auftreten, aber
    tmp = 1.0;
  }
  return (tmp/2.0);
}


// Um gueltig zu sein,
bool GesamtNetz::getGueltigkeit(void){
  if (gueltigkeit == 0)
    return gueltigkeit;

  // muessen diese Objekte existieren
  if (!Aouei || !Ouvoid || !Eivoid) {
    cerr << "WARNUNG: Wenigstens einer von Aouei, Ovoid und Eivoid == 0!\n";
    return 0;
  }

  // muessen die auch gueltig sein
  if (!Aouei->getGueltigkeit() || !Ouvoid->getGueltigkeit() || !Eivoid->getGueltigkeit())
    // das jeweilige Netz soll sich hier melden--deswegen kein explites
    // cerr hier.
    return 0;

  // muss die Magic-Number richtig sein
  if (strcmp(GESAMTNETZMAGICNUMBER,magicNumber)) {
    cerr << "WARNUNG: magicNumber ist nicht " << GESAMTNETZMAGICNUMBER << ", sondern " << magicNumber << ".\n";
    return 0;
  }

  return gueltigkeit;
}



//typedef enum {GesamtN, AoueiN, OuvoidN, EivoidN} WelchesNetz;
Statistik &GesamtNetz::getStatistik(WelchesNetz welches){
  if(getGueltigkeit()) {
    if (welches==GesamtN)
      return myStat;
    else if (welches==AoueiN)
      return Aouei->getStatistik();
    else if (welches==OuvoidN)
      return Ouvoid->getStatistik();
    else if (welches==EivoidN)
      return Eivoid->getStatistik();
  }

  cerr << "FEHLER: Problem in GesamtNetz::getStat(WelchesNetz)\n";
  cerr << "Problem mit WelchesNetz?  gueltigkeit des GesamtNetzes: " << getGueltigkeit() << "\n\n";
  exit(1);
}



GUS &GesamtNetz::getGewichtematrix(WelchesNetz welches, int schichtIndex){
  if(getGueltigkeit()){
    if (welches==GesamtN) {
      cerr << "FEHLER: Es gibt nur Gewichtematritze fuer die Expertennetze," << endl << "nicht fuer das GesamtNetz!\n";
      exit(1);
    }
    if (welches==AoueiN)
      return Aouei->getGewichtematrix(schichtIndex);
    else if (welches==OuvoidN)
      return Ouvoid->getGewichtematrix(schichtIndex);
    else if (welches==EivoidN)
      return Eivoid->getGewichtematrix(schichtIndex);
  }

  cerr << "FEHLER: getGewichtematrix() auf ein ungueltiges Gesamtnetz aufgerufen!\n";
  exit(1);
}


// Eine ganze Reihe von Settern und Gettern, die alle sehr aehnlich sind.
// Es wird angenommen, dass alle Expertennetze die gleichen Werten fuer
// epsilon usw. besitzen.
bool GesamtNetz::getTrainiert(void){
  if(getGueltigkeit())
    return Aouei->getTrainiert();
  
  cerr << "FEHLER: getTrainiert() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

void GesamtNetz::setTrainiert(bool isTrainiert){
  if(getGueltigkeit()){
    Aouei->setTrainiert(isTrainiert);
    Ouvoid->setTrainiert(isTrainiert);
    Eivoid->setTrainiert(isTrainiert);
    return;
  }

  cerr << "FEHLER: setTrainiert() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

int GesamtNetz::getBatchAnzahl(void){
  if(getGueltigkeit())
    return Aouei->getBatchAnzahl();
  
  cerr << "FEHLER: getBatchAnzahl() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

void GesamtNetz::setBatchAnzahl(int batchAnzahl){
  if(getGueltigkeit()){
    Aouei->setBatchAnzahl(batchAnzahl);
    Ouvoid->setBatchAnzahl(batchAnzahl);
    Eivoid->setBatchAnzahl(batchAnzahl);
    return;
  }

  cerr << "FEHLER: setBatchAnzahl() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

int GesamtNetz::getTrainingsfehlerAnzahl(void){
  if(getGueltigkeit())
    return Aouei->getTrainingsfehlerAnzahl();
  
  cerr << "FEHLER: getTrainingsfehlerAnzahl() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

void GesamtNetz::setTrainingsfehlerAnzahl(int trainingsfehlerAnzahl){
  if(getGueltigkeit()){
    Aouei->setTrainingsfehlerAnzahl(trainingsfehlerAnzahl);
    Ouvoid->setTrainingsfehlerAnzahl(trainingsfehlerAnzahl);
    Eivoid->setTrainingsfehlerAnzahl(trainingsfehlerAnzahl);
    return;
  }

  cerr << "FEHLER: setTrainingsfehlerAnzahl() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

int GesamtNetz::getTestfehlerAnzahl(void){
  if(getGueltigkeit())
    return Aouei->getTestfehlerAnzahl();
  
  cerr << "FEHLER: getTestfehlerAnzahl() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

void GesamtNetz::setTestfehlerAnzahl(int theTestfehlerAnzahl){
  if(getGueltigkeit()){
    Aouei->setTestfehlerAnzahl(theTestfehlerAnzahl);
    Ouvoid->setTestfehlerAnzahl(theTestfehlerAnzahl);
    Eivoid->setTestfehlerAnzahl(theTestfehlerAnzahl);
    return;
  }

  cerr << "FEHLER: setTestfehlerAnzahl() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

double GesamtNetz::getEpsilon(void){
  if(getGueltigkeit())
    return Aouei->getEpsilon();
  
  cerr << "FEHLER: getEpsilon() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}

void GesamtNetz::setEpsilon(double theEpsilon){
  if(getGueltigkeit()){
    Aouei->setEpsilon(theEpsilon);
    Ouvoid->setEpsilon(theEpsilon);
    Eivoid->setEpsilon(theEpsilon);
    return;
  }

  cerr << "FEHLER: setEpsilon() auf ein ungueltiges Netz aufgerufen!\n";
  exit(1);
}



int GesamtNetz::getSchichtenAnzahl(WelchesNetz welches){
  if(getGueltigkeit()){
    if (welches==GesamtN) {
      cerr << "FEHLER: Ein Gesamtnetz besitzt keine schichtenAnzahl!\n";
      exit(1);
    }
    if (welches==AoueiN)
      return Aouei->getSchichtenAnzahl();
    else if (welches==OuvoidN)
      return Ouvoid->getSchichtenAnzahl();
    else if (welches==EivoidN)
      return Eivoid->getSchichtenAnzahl();
  }

  cerr << "FEHLER: getSchichtenAnzahl() auf ein ungueltiges Gesamtnetz aufgerufen!\n";
  exit(1);
}



int *GesamtNetz::getTransferFunktionen(WelchesNetz welches){
  if(getGueltigkeit()){
    if (welches==GesamtN) {
      cerr << "FEHLER: Ein Gesamtnetz besitzt keine Transferfunktionen!\n";
      exit(1);
    }
    if (welches==AoueiN)
      return Aouei->getTransferFunktionen();
    else if (welches==OuvoidN)
      return Ouvoid->getTransferFunktionen();
    else if (welches==EivoidN)
      return Eivoid->getTransferFunktionen();
  }

  cerr << "FEHLER: getTransferFunktionen() auf ein ungueltiges Gesamtnetz aufgerufen!\n";
  exit(1);
}


int *GesamtNetz::getNeuronenAnzahl(WelchesNetz welches){
  if(getGueltigkeit()){
    if (welches==GesamtN) {
      cerr << "FEHLER: Ein Gesamtnetz besitzt keine neuronenAnzahlen!\n";
      exit(1);
    }
    if (welches==AoueiN)
      return Aouei->getNeuronenAnzahl();
    else if (welches==OuvoidN)
      return Ouvoid->getNeuronenAnzahl();
    else if (welches==EivoidN)
      return Eivoid->getNeuronenAnzahl();
  }

  cerr << "FEHLER: getNeuronenAnzahl() auf ein ungueltiges Gesamtnetz aufgerufen!\n";
  exit(1);
}

