#ifndef THEGESAMTNETZ
#define THEGESAMTNETZ

// Kann Control oder GUI bitte einmal die aktuelle Zeit zu dem 
// Zufallszahlengenerator liefern? (seed the random number generator)
#include "Schicht.h"
#include "Netz.h"
#include "Statistik.h"
#include "Vokal.h"

//(in Schicht.h definiert:)
//#define TANH 1
//#define LOGI 2
//#define IDENTITAET 0

#define GESAMTNETZMAGICNUMBER  "eDgR"    // ich speichere sie als char[4]
                                         // anstatt als ein int.
#define GESAMTNETZNOTMAGICNUMBER  "Invl"


typedef enum WelchesNetz {GesamtN, AoueiN, OuvoidN, EivoidN} WelchesNetz;

class GesamtNetz;    // forward declaration

ostream& operator<< (ostream &ostr, const GesamtNetz &);   // Allg. Ausgabe
istream& operator>> (istream &istr, GesamtNetz &);         // Allg. Eingabe

class GesamtNetz {
  friend ostream& operator<< (ostream &ostr, const GesamtNetz &);
  friend istream& operator>> (istream &istr, GesamtNetz &);

 protected:
  char magicNumber[5];    
  bool gueltigkeit;        // ob das GesamtNetz gueltig ist
  Statistik myStat;        // Die Statiken vom GesamtNetz selbst
  Vokal vok;
  // vok darf vom Aussen geaendert werden (wird manchmal als Vokal & 
  // zurueckgegeben), obwohl es eigentlich nicht sinnvoll waere... 
  // vok existiert hier nur damit es immer nur 
  // eine Kopie davon gibt (Speicherplatz sparen)...

  double testFehler, trainingsFehler;

  // Mehrere Variablen fuers berechnen:
  double AoueiSoll[3];
  double OuvoidSoll[3];
  double EivoidSoll[3];
  double *AoueiOut, *OuvoidOut, *EivoidOut;

  Netz *Aouei, *Ouvoid, *Eivoid;   // Zeiger auf die drei Expertennetze

 public:
  // Zunaechst ein Konstruktor, wobei der Parameter ein fstream & ist.
  // Das Gesamtnetz wird von der Datei geladen (drei Netze, Parameter, alte Statistiken)
  // Es wird von mir angenommen, dass fp schon geoeffnet wurde und in dem richtigen Modus ist.
  // Es wird auch angenommen, dass die Endianness schon richtig ist.
  // Mann kann Beispiele von GesamtNetzdateien mit makeNetz.cpp generieren...
  // Ihr sollt nachher fclose(fp) durchfuehren! Ich mache das nicht!
  // dh, usage:
  // fstream fp;
  // fp.open(filename,ios::in);
  // GesamtNetz g(fp);
  // fp.close();
  GesamtNetz(fstream &fp);

  // Und dann ein Konstruktor, der ein untrainiertes Netz schoepft.
  // Die meisten von diesen Parametern koennen nicht spaeter geandert werden, weil
  // es zum Beispiel bedeutungslos waere, ein paar neuen Schichten mit untrainierten
  // Gewichten/Schwellen zu einem trainierten Netz hinzuzufuegen.
  //
  // 1 ist fuer das A-O/U-E/I Netz, 2 fuer O-U-void und 3 fuer E-I-void 
  //   schichtenAnzahl: wie viele Schichte es besitzt (wenigstens 2, sinnvoll sind 3+,
  //                    und wahrscheinlich nicht mehr als 10)
  //   tf: welche TransferFunktion pro Schicht benutzt werden soll (TANH, LOGI oder IDENTITAET).
  //       tf[0] (fuer die nullte Schicht) ist egal, dh. tf[0] == IDENTITAET; wird immer von mir
  //       durchgefuehrt
  //   neuronenAnzahl: wie viele Neuronen es pro Schicht geben soll
  //                   *Achtung! Die Neuronenanzahl der 0. Schicht MUSS == die Anzahl der Fenster von
  //                   *der auditorischen Vorverarbeitung!
  //                   *Und die Neuronenanzahl der letzten Schicht MUSS == 3 sein!
  // anschliessend auch einmal (diese koennen spaeter mit Gettern und Settern geaendert werden):
  //   batchAnzahl: ob Online-learning benutzt werden soll (==1) oder die Anzahl der Schritte
  //                zwischen Updates == die Laenge der Epoche (S. 20)
  //   trainingsfehlerAnzahl: wie viele Trainingsfiles angekuckt werden, vor ein neuer Datenpunkt
  //                          zu den Trainierungsstatiken zugefuegt werden soll. (ankucken: trainiere)
  //   testfehlerAnzahl:      wie viele Testfiles angekuckt werden, vor ein neuer Datenpunkt
  //                          zu den Teststatiken zugefuegt werden soll. (ankucken: test)
  //   oft ist batchAnzahl == trainingsFehlerAnzahl == testFehlerAnzahl
  //   sinnvoll ist trainingsFehlerAnzahl == testFehlerAnzahl, obwohl nicht zwingend (Plots saehen
  //   komisch aus..)
  //   und immer ist batchAnzahl == Laenge der Epoche
  //   epsilon: siehe S. 20, im Bereich [10^-3 0.5]
  // Alle dieser Parameter sollen vom Benutzer einstellbar sein!
  GesamtNetz(int schichtenAnzahl1, int *tf1, int *neuronenAnzahl1,
	     int schichtenAnzahl2, int *tf2, int *neuronenAnzahl2,
	     int schichtenAnzahl3, int *tf3, int *neuronenAnzahl3,\
	     int batchAnzahl, int trainingsfehlerAnzahl, 
	     int testfehlerAnzahl, double epsilon);

  // Leerer Konstruktor, damit man ein Gesamtnetz machen kann und spaeter
  // durch >> einstellen.
  // Ich wuerde ihn nicht so sehr empfehlen, da ich ihn nicht so gut getestet habe,
  // und ihr braucht ihn wahrscheinlich auch nicht.
  GesamtNetz(void);

  // Das Gesamtnetz wird in dieser Datei gespeichert, damit es spaeter durch GesamtNetz(char *)
  // geladen werden kann. Es wird angenommen, dass fp schon geoeffnet wurde und in dem richtigen
  // Modus ist (und leer ist). 
  // Ihr sollt nacher fclose(fp) durchfuehren! Ich mache das wieder nicht!
  // dh, usage:
  // fstream fp;
  // fp.open(filename,ios::out);
  // g.speichern(fp);
  // fp.close();
  bool speichern(fstream &fp);  

  // Trainiere das Gesamtnetz mit der Ausgabe von der Vorverarbeitung.
  // Die Laenge von FFTFenster muss gleich der Dimension (Anzahl der 
  // Neuronen==outputDim) der nullte Schicht in allen drei Netzen sein.
  // Vokal.gewaehlt und Vokal.sicherheit koennen falsch sein: welches 
  // ebenso wahrscheinlicher ist, wenn die Netze noch nicht gut trainiert sind.
  // Siehe die Definition von Vokal oben.
  // vokl.neuesFehlerplot und vokl.neueGewichtenwerte sollen JEDES MAL 
  // ueberprueft werden, um zu sehen, ob das Plot geupdatet werden soll.
  // Hinweis: Diese Fehler werden aber hier nicht zurueckgegeben. Sie
  // muessen durch getStat() zugegriffen werden.
  // Beispiel usage: Vokal &vokl = g.trainiere(FFTFenster, 'a');
  //   (wobei g ein gueltiges GesamtNetz ist)
  // Siehe Vokal.h fuer mehr Details
  Vokal &trainiere(double *FFTFenster, char echterVokal);

  // Das gleiche wie trainiere, aber die Schwellen und Gewichten werden nicht geaendert.
  // Die Testfehler anstatt der Trainingsfehler werden natuerlich hier berechnet.
  // Hinweis: Diese Fehler werden aber hier nicht zurueckgegeben. Sie 
  // muessen durch getStat() zugegriffen werden.
  Vokal &test(double *FFTFenster, char echterVokal);

  // Eingabe wird durch die Netze geschickt und Ausgabe ausgegeben.
  // Allerdings werden keine Fehler berechnet und keine Parameter der Netze
  // gaendert.
  // Trotzdem ist es wahrscheinlich eine gute Idee, vokl.neuesFehlerplot
  // und vokl.neueGewichtenwerte zu ueberpruefen.
  Vokal &propagate(double *FFTFenster);

 private:   // einige Hilfefunktionen:
  void setSoll(char echterVokal);
  void vomNetzenPseudonormieren(void);
  void outputPseudonormieren(void);
  double vokalWert(void);
  void charWaehlen(void);
  double gesamtfehlerBerechnen(char);  
 public:
  // Getter/Setter:
  // zeigt ob das Netz gueltig ist oder nicht
  // (beim Einlesen eines falschen Gesamtnetzes kann dieser Fall zB auftreten)
  // getGueltigkeit() wird auch ueberprueft vor ein Netz gespeichert werden kann.
  // Diese Funktion soll aber nicht oft aufgerufen werden. Ich habe sie nicht
  // so gut getestet, und sie dauert ein bisschen.
  bool getGueltigkeit(void);

  // Damit ihr die Plots darstellen koennt (wie auf S. 13)
  // usage: Statistik &stat = getStatistik(AoueiN);
  Statistik &getStatistik(WelchesNetz);
  
  // Um die Werte fuer ein Plot wie auf S. 21 zu bekommen
  // Fuer ein Netz mit 3 Schichten (wie auf S. 16) ist die 0. Schichte trivial aber
  // trotzdem darstellbar. Die 1. und 2. Schichten sind mehr interessant.
  // usage Beispiel: GUS &darzustellen = getGewichtematrix(EivoidN, 1);
  // Siehe GUS.h
  GUS &getGewichtematrix(WelchesNetz welches, int schichtIndex);

  // sehr einfache Getter und Setter
  bool getTrainiert(void);               // zeigt ob das Netz trainiert ist
  void setTrainiert(bool isTrainiert);   // Control/GUI entscheidet, wenn das Netz trainiert ist
  int getBatchAnzahl(void);              // siehe die Definition des Konstruktors oben
  void setBatchAnzahl(int batchAnzahl);
  int getTrainingsfehlerAnzahl(void);
  void setTrainingsfehlerAnzahl(int trainingsfehlerAnzahl);
  int getTestfehlerAnzahl(void);
  void setTestfehlerAnzahl(int theTestfehlerAnzahl);
  double getEpsilon(void);
  void setEpsilon(double theEpsilon);
  
  // Neu fuer die Oberflaeche:
  // Fuer die gibt es keine Setter, weil ein neues Netz sowieso generiert werden muss, falls diese
  // Parameter geaendert werden sollen.
  int getSchichtenAnzahl(WelchesNetz);        // Anzahl der Schichten in einem Netz (wenigstens 2!)
  int *getTransferFunktionen(WelchesNetz);    // die Laenge des Ergebnisses ist schichtenAnzahl
  int *getNeuronenAnzahl(WelchesNetz);        // die Laenge des Ergebnisses ist schichtenAnzahl
};

#endif
