#include "Neuron.h"
#include "stdlib.h"       // fuer rand()
//#define AUSGABELESBAR   // damit Menschen die << Ausgabe von Neuronen lesen koennen

 
// Fuer allgemeine Ausgabe und Eingabe
// muss nachher eingestellt werden!
// Format ist jetzt: [inputDim] [gewicht_0 gewicht_1 ... gewicht_(inputDim-1)] [schwelle] <CR>
// Fuers Testen, damit ich cout benutzen kann
ostream& operator<< (ostream &ostr, const Neuron &theNeuron){

  // mit Eingabe kompatibel
#ifndef AUSGABELESBAR
  ostr << theNeuron.inputDim << " ";
  if(theNeuron.inputDim<2) {
    if(theNeuron.inputDim==1) 
      ostr << theNeuron.gewichte[0];
  } else {
  for(int i=0;i<(theNeuron.inputDim-1);i++)
    ostr << theNeuron.gewichte[i] << " ";
  ostr << theNeuron.gewichte[theNeuron.inputDim-1];
  }
  ostr << " " << theNeuron.schwelle << flush;
#endif


  // aber fuer Menschen besser lesbar
#ifdef AUSGABELESBAR
  ostr << "{" << theNeuron.inputDim << ":(";
  if(theNeuron.inputDim<2) {
    if(theNeuron.inputDim==1) 
      ostr << theNeuron.gewichte[0];
  } else {
  for(int i=0;i<(theNeuron.inputDim-1);i++)
    ostr << theNeuron.gewichte[i] << ",";
  ostr << theNeuron.gewichte[theNeuron.inputDim-1];
  }
  ostr << ")," << theNeuron.schwelle << flush;
#endif

  return(ostr);
}


// Fuers Testen, und vielleicht auch fuers Laden einer Netzdatei
istream& operator>> (istream &istr, Neuron &theNeuron){
  istr >> theNeuron.inputDim;

  if(theNeuron.inputDim > 0) {
    if(theNeuron.gewichte)
      delete[] theNeuron.gewichte;
    theNeuron.gewichte = new double[theNeuron.inputDim];

    if(!theNeuron.gewichte) {
      cerr << "Allokationsfehler" << endl;
      exit(1);
    }

    for(int i=0;i<theNeuron.inputDim;i++)
      istr >> theNeuron.gewichte[i];
    istr >> theNeuron.schwelle;
  }

  return istr;
}



// Konstruktor, der alles mit Zufallswerten initialisiert 
// In dem Bereich [-RANDSKALIERUNG,RANDSKALIERUNG]
Neuron::Neuron(int theInputDim):inputDim(theInputDim){
  deltaGewichte = 0;        // noch nicht initialisiert!
  deltaSchwelle = 0.0;
  gewichte = 0;             // falls zB theInputDim == -1

  if (theInputDim >= 0) {
    gewichte = new double[inputDim];
    
    if(!gewichte) {
      cerr << "Allokationsfehler" << endl;
      exit(1);
    }
    
    for(int i=0;i<inputDim;i++)
      gewichte[i] = Neuron::doRand();
  }
  schwelle = Neuron::doRand();
}



// Konstruktor, der alles mit der Eingabe initialisiert
// zB gut geeignet fuer die 0. Schicht, wo die Gewichten [0,1,0,0] sind
Neuron::Neuron(int theInputDim, double *theGewichte, 
	       double theSchwelle):inputDim(theInputDim){
  deltaGewichte = 0;        // noch nicht initialisiert!
  deltaSchwelle = 0.0;
  gewichte = 0;             // falls zB theInputDim == -1

  if (theInputDim >= 0) {
    gewichte = new double[inputDim];
    
    if(!gewichte) {
      cerr << "Allokationsfehler" << endl;
      exit(1);
    }
   
    for(int i=0; i<inputDim; i++)
      gewichte[i] = theGewichte[i];
    schwelle = theSchwelle;
  }
}



// Destruktor, um Memory-Leaks zu vermeiden
 Neuron::~Neuron(void){
   //cout << "gewichte:" << gewichte << ", deltaGewichte: " << deltaGewichte << endl << flush;
   if(gewichte)
     delete[] gewichte;
   if(deltaGewichte)
     delete[] deltaGewichte;
 }


// Copy-Konstruktor
// deltaGewichte[], deltaSchwelle werden nicht kopiert bzw. nicht initialisert!
Neuron::Neuron(const Neuron &source){
  inputDim = source.inputDim;
  schwelle = source.schwelle;

  if(inputDim>=0) {
    gewichte = new double[inputDim];
    if(!gewichte) {
      cerr << "Allokationsfehler" << endl;
      exit(1);
    }
    if(source.gewichte)
      for(int i=0;i<inputDim;i++)
	gewichte[i] = source.gewichte[i];
  }

  deltaGewichte = 0;
  deltaSchwelle = 0.0;
}




// Fuehrt die erste Haelfe der Forward-Propagation fuer dieses Neuron durch
// die Transferfunktion muss nachher angewendet werden!
// Ich mache das nicht hier, weil die Transferfunktion eigentlich zu der Schicht gehoert
// Es wird angenommen, dass die Dimensionen uebereinstimmen!
double Neuron::propagate(double *input){
  double summe = 0.0;

  for(int i=0;i<inputDim;i++)
    summe += input[i]*gewichte[i];

  return (summe-schwelle);
}


// Ergebnis ist ein fuer den Anfang geeigneter Wert im Bereich [-0.1,0.1]
double Neuron::doRand(void){
  return (((double)rand())/((double)RAND_MAX)-0.5)*2.0*RANDSKALIERUNG;
}
  

// Die Gewichte und die Schwelle werden jetzt nach deltaGewichte und deltaSchwelle geaendert 
// deltaGewichte[] und deltaSchwelle werden dann zu 0 gesetzt
// Diese Funktion wird fuer Backpropagation benutzt
void Neuron::doDeltas(void){
  if(!deltaGewichte || !gewichte || (inputDim<1)) {
    cerr << "WARNUNG: Versuch doDeltas() durchzufuehren, vor alles initialisiert wurde!" << endl << flush;
  } else {
    for(int i=0;i<inputDim;i++) {
      gewichte[i] += deltaGewichte[i];
      deltaGewichte[i] = 0.0;
    }
    schwelle += deltaSchwelle;
    deltaSchwelle = 0.0;
  }
}


// Man kann Aenderungen in deltaGewichte[] und deltaSchwelle hier eingeben
// Diese Funktion wird fuer Backpropagation benutzt
// Falls deltaGewichte[] noch nicht existiert, wird es hier initialisert
// (Dh, wird nur initialisiert, wenn es gebraucht wird)
// (deltaGewichte[] deltaSchwelle koennen nur zu 0 durch doDeltas() gesetzt)
void Neuron::updateDeltas(double *theDeltaGewichte, double theDeltaSchwelle){
  if(!deltaGewichte) {
    deltaGewichte = new double[inputDim];
    if(!deltaGewichte) {
      cerr << "Allokationsfehler!";
      exit(1);
    } else {
      for(int i=0;i<inputDim;i++)
	deltaGewichte[i] = 0.0;
      deltaSchwelle = 0.0;
    }
  }

  for(int i=0;i<inputDim;i++)
    deltaGewichte[i] = deltaGewichte[i] + theDeltaGewichte[i];
  deltaSchwelle = deltaSchwelle + theDeltaSchwelle;
}
