#include "Schicht.h"
#include <iostream>
#include <math.h>
 
// Fuer allgemeine Ausgabe und Eingabe
// muss nachher eingestellt werden!
// Format ist jetzt: [outputDim]\t[inputDim]\t[tFTyp]
//                   [Neuron_0]
//                   [Neuron_1] ...
//                   [Neuron_(outputDim-1)]
// mit Eingabe kompatibel!
ostream& operator<< (ostream &ostr, const Schicht &theSchicht){
  ostr << theSchicht.outputDim << "\t" << theSchicht.inputDim
       << "\t" << theSchicht.tFTyp;
  for(int i=0;i<theSchicht.outputDim;i++) 
    ostr << endl << *(theSchicht.Neurons[i]);
  ostr << endl << flush;
  
  return(ostr);
}


// Fuers Testen, und vielleicht auch fuers Laden einer Netzdatei
istream& operator>> (istream &istr, Schicht &theSchicht){
  int oldOutputDim = theSchicht.outputDim;

  istr >> theSchicht.outputDim;
  istr >> theSchicht.inputDim;
  istr >> theSchicht.tFTyp;

  //remove this line later!
  //theSchicht.Neurons = 0;

  if ((theSchicht.outputDim >= 1) && (theSchicht.inputDim >= 1)) {
    if(theSchicht.Neurons) {
      for(int i=0;i<oldOutputDim;i++)
	if (theSchicht.Neurons[i])
	  delete theSchicht.Neurons[i];
      delete[] theSchicht.Neurons;
    }

    theSchicht.Neurons = new (Neuron *)[theSchicht.outputDim];
    if (!theSchicht.Neurons) {
      cerr << "Allokationsfehler!" << endl;
      exit(1);
    }
    for(int i=0;i<theSchicht.outputDim;i++) {
      theSchicht.Neurons[i] = new Neuron(theSchicht.inputDim);
      if (!theSchicht.Neurons[i]) {
	cerr << "Allokationsfehler!" << endl;
	exit(1);
      }
    }
    for(int i=0;i<theSchicht.outputDim;i++)
      istr >> *(theSchicht.Neurons[i]);
  } else {
    cerr << "Dimensionen fuer Schicht::Schicht() muessen beide groesser als 0 sein!";
    exit(1);
  }

  return istr;
}

// Bemerkung: inputDim soll == Neurons[x].inputDim, wobei 0<=x<outputDim
Schicht::Schicht(int theInputDim, int theOutputDim, 
		 int theTFTyp):inputDim(theInputDim),outputDim(theOutputDim),
			       Neurons(0),
			       tFTyp(theTFTyp),propagateAusgabe(0),
			       getSchwellenAusgabe(0),getGewichteAusgabe(0){
  if ((outputDim >= 1) && (inputDim >= 1)) {
    Neurons = new (Neuron *)[outputDim];
    if (!Neurons) {
      cerr << "Allokationsfehler!" << endl;
      exit(1);
    }
    for(int i=0;i<outputDim;i++) {
      Neurons[i] = new Neuron(inputDim);
      if (!Neurons[i]) {
	cerr << "Allokationsfehler!" << endl;
	exit(1);
      }
    }      
  } else {
    cerr << "Dimensionen fuer Schicht::Schicht() muessen beide groesser als 0 sein!";
    exit(1);
  } 
}


Schicht::Schicht(int theInputDim, int theOutputDim, 
		 int theTFTyp, double **gewichte, 
		 double *schwellen):inputDim(theInputDim),outputDim(theOutputDim),Neurons(0),
				    tFTyp(theTFTyp),propagateAusgabe(0),
				    getSchwellenAusgabe(0),getGewichteAusgabe(0){
  if ((outputDim >= 1) && (inputDim >= 1)) {
    Neurons = new (Neuron *)[outputDim];
    if (!Neurons) {
      cerr << "Allokationsfehler!" << endl;
      exit(1);
    }
    for(int i=0;i<outputDim;i++) {
      Neurons[i] = new Neuron(inputDim,gewichte[i],schwellen[i]);
      if (!Neurons[i]) {
	cerr << "Allokationsfehler!" << endl;
	exit(1);
      }
    }      
  } else {
    cerr << "Dimensionen fuer Schicht::Schicht() muessen beide groesser als 0 sein!";
    exit(1);
  }
}


// ein Konstruktor, der besonders gut fuer die nullte Schicht geeignet ist
Schicht::Schicht(int theInputDim, int theOutputDim, int theTFTyp, bool dummy):inputDim(theInputDim),outputDim(theOutputDim),Neurons(0),tFTyp(theTFTyp),propagateAusgabe(0),getSchwellenAusgabe(0),getGewichteAusgabe(0){
  dummy = 0;    // damit der Compiler nicht meckert
  
  if ((theInputDim!=theOutputDim) || (theInputDim < 0) || (theOutputDim < 0)) {
    cerr << "Dimensionierungsfehler in Schicht fuer die 0. Schicht!";
    exit(1);
  }

  double **identitaet = new (double *)[theInputDim];
  if (!identitaet) {
    cerr << "Allokationsfehler!";
    exit(1);
  }

  for(int i=0;i<theInputDim;i++){
    identitaet[i] = new double[theInputDim];
    if (!identitaet[i]) {
      cerr << "Allokationsfehler!";
      exit(1);
    }
  }

  double *schw = new double[theInputDim];
  if (!schw) {
    cerr << "Allokationsfehler!";
    exit(1);
  }

  for(int j=0;j<theInputDim;j++) {
    for(int k=0;k<theInputDim;k++) {
      if (j==k)
	identitaet[j][k] = 1.0;
      else
	identitaet[j][k] = 0.0;
    }
    schw[j] = 0.0;
  }

  //Neurons = 0;
  
  if ((outputDim >= 1) && (inputDim >= 1)) {
    Neurons = new (Neuron *)[outputDim];
    if (!Neurons) {
      cerr << "Allokationsfehler!" << endl;
      exit(1);
    }
    for(int i=0;i<outputDim;i++) {
      Neurons[i] = new Neuron(inputDim,identitaet[i],schw[i]);
      if (!Neurons[i]) {
	cerr << "Allokationsfehler!" << endl;
	exit(1);
      }
    }      
  } else {
    cerr << "Dimensionen fuer Schicht::Schicht() muessen beide groesser als 0 sein!";
    exit(1);
  }
}


Schicht::~Schicht(void){
  if(propagateAusgabe)
    delete[] propagateAusgabe;
  if(getSchwellenAusgabe)
    delete[] getSchwellenAusgabe;

  if (Neurons) {
    for(int i=0;i<outputDim;i++)
      if (Neurons[i])
	delete Neurons[i];
    delete[] Neurons;
  }

  // Obwohl dies ist ein double **, muss nur vom Aussen geloescht werden.
  // Die inneren double *'s sind mit den Neuronen geloeschen worden!
  // (Das hat Segmentierungsfehler verursacht, aber nur bei Linux.)
  if(getGewichteAusgabe)
    delete[] getGewichteAusgabe; 
}


double *Schicht::propagate(double *input){
  if(!propagateAusgabe) {
    propagateAusgabe = new double[outputDim];
    if(!propagateAusgabe) {
      cerr << "Allokationsfehler!";
      exit(1);
    }
  }
  
  double temp;
  
  for(int i=0;i<outputDim;i++) {
    temp = Neurons[i]->propagate(input);
    
    if (tFTyp==TANH) {
      temp = tanh(temp);
    } else if (tFTyp == LOGI) {
      temp = 1.0/(1.0 + exp((-1.0) * temp));
    } // else, keine Veraenderung (IDENTITAET)
    
    propagateAusgabe[i] = temp;
  }

  return(propagateAusgabe);
}


void Schicht::doDeltas(void){
  for(int i=0;i<outputDim;i++)
    Neurons[i]->doDeltas();
}


double *Schicht::recentPropagate(void){
  return propagateAusgabe;
}


void Schicht::updateDeltas(double **gewichte, double *schwellen){
  for(int i=0;i<outputDim;i++)
    Neurons[i]->updateDeltas(gewichte[i], schwellen[i]);
}


double *Schicht::getSchwellen(void){
  if(!getSchwellenAusgabe){
    getSchwellenAusgabe = new double[outputDim];
    if (!getSchwellenAusgabe) {
      cerr << "Allokationsfeh  \n";
      exit(1);
    }    
  }

  for(int i=0;i<outputDim;i++)
    getSchwellenAusgabe[i] = Neurons[i]->getSchwelle();

  return getSchwellenAusgabe;
}


double **Schicht::getGewichte(void){
  if(!getGewichteAusgabe){
    getGewichteAusgabe = new (double *)[outputDim];
    if (!getGewichteAusgabe) {
      cerr << "Allokationsfehler!\n";
      exit(1);
    }
  }

  for (int i=0;i<outputDim;i++)
    getGewichteAusgabe[i] = Neurons[i]->getGewichte();

  return getGewichteAusgabe;
}

