#include <string>

using namespace std;

/*
	Klasse, die der Verwaltung von *.set Dateien dient

	Eine *.set Datei ist eine ASCII-Datei, die die absoluten Pfade von mehreren *.au Dateien
	enthlt. Dies dient der besseren Verwaltung von greren Test-und Trainingsdatenstzen.


	Aufbau einer *.set Datei:

	In der ersten (0-basiert) Zeile steht die Anzahl der aktuell enthaltenen Pfade.
	Dann folgt in jeder Zeile genau ein Pfad.
	Die Reihenfolge der Pfade ist alphabethisch.
	Die Datei wird mit newline abgeschlossen.
	Die erste un die letzte Zeile sind fuer den User nicht beschreibbar.
*/


class SetFile
{

	public:

		// leerer Konstruktor
		SetFile();

		// Konstruktor mit Pfadnamen der *.set Datei
		// *.set Datei wird erstellt, und 0 in die 1. Zeile geschrieben
		// wenn diese Datei schon existiert, wird sie ueberschrieben
		SetFile(string setfilename);

		// Destruktor
		~SetFile();

		// setzt SetFileName, also den Namen der *.set Datei, die verwaltet werden soll
		// wenn die Datei schon existiert, werden ihre Daten uebernommen,
		// ansonsten wird die Datei neu angelegt und 0 in die 1. Zeile geschrieben
		void setFileName(string setfilename);

		// fgt den Pfad einer *.au Datei in der *.set Datei ein;
		// der Pfad wird an der richtigen Stelle einsortiert
		// ist der Eintrag aber schon vorhanden, wird der Pfad nicht eingefuegt (return = 0)
		// Rckgabewert ist -1 bei Fehler, sonst Zeilennummer
		// weitere Erklaerungen siehe add(int, string) unter private
		int add(string aufilename);

		// lscht den Pfad einer *.au Datei aus der *.set Datei
		// Rckgabewert ist -1 bei Fehler, sonst Zeilennummer
		int rem(string aufilename);

		// lscht Pfad in bestimmter Zeile der *.set Datei
		// 0 < zeile <= dim+1
		// Rckgabewert ist -1 bei Fehler
		int rem(int zeile);

		// liefert die i-ten Zeile, 0 <= i <= dim+1
		// liefert "" bei Fehler
		// fuer i==0 wird die Dimensionsangabe der Datei geliefert, also kein Pfad
		string getPath(int i);

		// liefert den Pfad, auf den Zeiger ptr gerade weist
		// und zhlt dann den Zeiger auf den nchsten Pfad hoch
		// liefert "" bei Fehler
		string getPath();

		// liefert die Anzahl der Pfade in der Datei
		int getDim(void){ return dim; }

		// liefert die Zeile, auf die ptr aktuell zeigt
		int getPtr() { return ptr; }

		// setzt den Zeiger ptr auf eine bestimmte Zeile
		// Rckgabewert ist -1 bei Fehler (z.B. wenn er auf nicht existierende Zeile gesetzt wird)
		// sonst wird Zeile zurueckgegeben auf die ptr gesetzt wurde
		int setPtr(int i);

		// setzt den Zeiger ptr auf 1. Datenzeile, d.h. 2.Zeile der Datei
		void reset() { ptr = 1; }

	private:

		// gibt an, wieviele Pfade die *.set Datei enthlt
		// entspricht der 1. Zeile der *.set Datei
		int dim;

		// der Name der *.set Datei, die verwaltet wird
		string FileName;

		// Zeiger auf die aktuelle Zeile
		// wird immer mit 1 initialisiert, zeigt also am Anfang
		// immer auf 2. Zeile der Datei das ist die erste Datenzeile
		int ptr;

		// fgt Pfad in bestimmter Zeile der *.set Datei ein
		// der Rest rutscht eins weiter nach unten (die Zeile wird also nicht berschrieben)
		// Rckgabewert ist -1 bei Fehler
		// die *.set Datei muss schon existieren, sie wird
		// an dieser Stelle nicht angelegt
		// Man kann erst ab der 2. (0-basiert) Zeile schreiben,
		// da die erste Zeile fuer Dimension reserviert ist
		// d.h. zeile >= 1
		// es werden keine zusaetzliche Leerzeilen eingefuegt, d.h. 
		// zeile <= dim+1 (entspricht letzter Zeile)
		int add(int zeile, string aufilename);
		
		// wandelt String in int um
		// liefert -1 bei Fehler sonst 0
		int string2int(char* digit, int& result);

};
